/*-------------------------------------------------
 *    Kernel version 2.4.20 - 2.6.8, 2.6.13
 *    RS485 (COM-port)
 * 
 *    Andre Zaitsev, Olga Voziakova
 *    ovoz@sai.msu.ru
 *    $History:
 *
 *---------------------------------------- */
#include <linux/module.h>
#include <linux/types.h>
//#include <linux/kdev_t.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <mcua.h>         // To define KERN_2_6
#ifndef KERN_2_6_18
   #include <linux/devfs_fs_kernel.h>
#endif


#include <linux/errno.h>
#include <linux/slab.h>
#include <asm/uaccess.h>    //  copy_from_user
#include <linux/ioport.h>
#include <linux/delay.h>    // udelay
#include <linux/interrupt.h>
#include <asm/irq.h>
#include <asm/io.h>

#include <com.h>
#include <rs485.h>
#include <rs485_tools.h>
#include <rs485_irq.h> 
#include <rserrno.h>
#include <rs485_fifo.h>

static unsigned int major;

#define COM_MAX 2
RS485INFO com_table[COM_MAX];
static int __initdata io[COM_MAX] = {_IO_COM1, [1 ... COM_MAX-1]=_IO_COM2 };
static int __initdata irq[COM_MAX] = {_IRQ_COM1, [1 ... COM_MAX-1]=_IRQ_COM2};
static int __initdata baudrate[COM_MAX] = {_BAUDRATE_1, [1 ... COM_MAX-1]=_BAUDRATE_2};

struct file_operations rs485_fops;

rs485_time_t end_packet_out_tm;
rs485_time_t start_packet_inp_tm;
rs485_time_t tm;

inline int rs485_getIO( int port) {
    return ( port==2 )? _IO_COM2 : _IO_COM1;
}

inline int rs485_getIRQ( int port ) {
    return ( port==2 )? _IRQ_COM2 : _IRQ_COM1;
}  

//inline int rs485_get_port(int irq) {
//    return ( irq == irq1 )? 1 : 2;
//}


inline void rs485_allow_interrupt_card(RS485INFO *pinfo) {
    outb( UART_IER_RDI, pinfo->base +UART_IER );
}

inline void rs485_stop_interrupt_card(RS485INFO *pinfo) {
    outb( 0, pinfo->base +UART_IER );
}

inline void rs485_transmit_disable(RS485INFO *pinfo) {
    outb( RXENABLE, pinfo->base + RX_TX_REG );
}

inline void rs485_transmit_enable(RS485INFO *pinfo) {
    outb( TXENABLE, pinfo->base + RX_TX_REG );
}

int rs485_test_interface(RS485INFO *pinfo) {
    rs485_transmit_disable(pinfo); // ?? 
    if( ( inb( pinfo->base + UART_MSR ) & UART_MSR_CTS ) == 0 ) return -ERSNOINTERFACE; 
    return SUCCESS;
}

int rs485_test_power( RS485INFO *pinfo ) {
    rs485_stop_interrupt_card( pinfo );
    rs485_transmit_enable( pinfo );
    inb_p(pinfo->base);
    outb_p( 0x55, pinfo->base );  // Test byte
    rs485_set_starttime();
    while( ( inb_p( pinfo->base + UART_LSR ) & UART_LSR_DR ) == 0 ) {
        if( rs485_timeout( TBYTE_TIMEOUT ) ) {
            rs485_transmit_disable(pinfo);
            rs485_allow_interrupt_card( pinfo );
	    PDEBUG( "test_power: timeout: no echo!\n" );
	    return -ERSSHUTDOWN;   // no echo => no power
        }
        udelay( TBYTE_WAIT );
    }
    if( inb_p(pinfo->base) != 0x55 ) PDEBUG( "rs485_test_power: BAD ECHO\n");
    rs485_transmit_disable( pinfo );
    rs485_allow_interrupt_card( pinfo );
    return SUCCESS;
}

int rs485_test_device(RS485INFO *pinfo) {
    int ret = SUCCESS;
    if( ( ret = rs485_test_interface(pinfo) ) != SUCCESS ) return ret;
    return rs485_test_power( pinfo );
}

static unsigned char uart_ier_back = 0;
static unsigned char uart_lcr_back = 0;
static unsigned char uart_mcr_back = 0;

void rs485_get_reg(RS485INFO *pinfo) {
    uart_ier_back = inb_p(pinfo->base +UART_IER);
    uart_lcr_back = inb_p(pinfo->base +UART_LCR);
    uart_mcr_back = inb_p(pinfo->base +UART_MCR);
}

void rs485_restore_reg(RS485INFO *pinfo) {
    outb_p(uart_ier_back, pinfo->base +UART_IER);
    outb_p(uart_lcr_back, pinfo->base +UART_LCR);
    outb_p(uart_mcr_back, pinfo->base +UART_MCR);
}

int ubrr_to_rate(int baudrate_code) {
    if( baudrate_code == 0 ) return 0;
    return(Fosc/16/baudrate_code);
}

int rate_to_ubrr(int baudrate) {
    int baudrate_code;
    baudrate = baudrate < Minbaudrate ? Minbaudrate : baudrate;
    baudrate = baudrate > Maxbaudrate ? Maxbaudrate : baudrate;
    baudrate_code = Fosc/16/baudrate;
    return baudrate_code;
}

int rs485_set_baudrate(RS485INFO *pinfo) {
    int baudrate_code;
    baudrate_code = rate_to_ubrr( pinfo->baudrate );
    pinfo->baudrate = ubrr_to_rate( baudrate_code );
    PDEBUG( "set baudrate=%d\n", pinfo->baudrate );
    outb_p( 0x80, pinfo->base + UART_LCR );
    outw( baudrate_code, pinfo->base +UART_DLL ); 
    outb_p( OTHER, pinfo->base + UART_LCR );
    return SUCCESS;
}

int rs485_reset(RS485INFO *pinfo) {
    rs485_transmit_disable(pinfo);
    msleep( 8 );
    outb_p( UART_FCR_ENABLE_FIFO|UART_FCR_CLEAR_RCVR|UART_FCR_CLEAR_XMIT|UART_FCR_TRIGGER_1, pinfo->base+UART_FCR );
    inb_p( pinfo->base+UART_RX );
    rs485_set_baudrate( pinfo );
//    rs485_allow_interrupt_card( pinfo );
    return rs485_test_device( pinfo );
}

// -----------------------------------------------------------------------

inline int getbyte( RS485INFO *info, unsigned char *b, int resp_err ) {
    unsigned short iir;
    *b = 0;
    rs485_set_starttime();
    while( (( iir = inb( info->base+UART_LSR ) ) & UART_LSR_DR ) == 0 ) {
        if( rs485_timeout( TBYTE_TIMEOUT ) ) return resp_err;
        udelay( TBYTE_WAIT );
    }
    *b = inb_p( info->base );        
    if( ( iir & UART_LSR_FE ) != 0 ) {
        PDEBUG( "transmit: frame error!\n" );
        return SignalEER;
    }
    return 0; // OK
}

static unsigned char echo=0;
static int ret = 0;
static volatile unsigned char signal_from_device = 0;

#ifdef __DEBUG_TIMING__
    static rs485_time_t tm_start;
#endif

// debug
static unsigned char last_address=0;

inline void cleanFIFO(RS485INFO *info) {
    outb_p( UART_FCR_ENABLE_FIFO|UART_FCR_CLEAR_RCVR|UART_FCR_CLEAR_XMIT|UART_FCR_TRIGGER_1, info->base+UART_FCR );
    inb( info->base );
}

int start_transmitter( RS485INFO *info ) {
    unsigned int i;
    cleanFIFO(info);
    rs485_stop_interrupt_card(info);
    rs485_stop_interrupts();
    info->rs->status = PASSIVE_DATA;  // it means that PASSIVE_DATA will be wait

// TRANSMITTER:
    outb_p(FIRST, info->base + UART_LCR);
    rs485_transmit_enable(info);
    
    last_address = info->rs->outbuffer.cache[0];
    for(i = 0; i <= info->rs->outbuffer.size; i++) {
        outb_p( info->rs->outbuffer.cache[i], info->base);
        ret = getbyte(info, &echo, -ERSSHUTDOWN);
        if( i == 0 ) outb_p( OTHER, info->base + UART_LCR ); // address-byte
        if( ret == 0 && echo != info->rs->outbuffer.cache[i] ) ret = SignalCOL;
        if( ret != 0 ) {
            PDEBUG("addr=%hx transmit: getbyte timeout, i=%d, info->rs->outbuffer.size=%zd\n",
                                last_address, i, info->rs->outbuffer.size);
            rs485_transmit_disable(info);
            rs485_allow_interrupts();
            rs485_allow_interrupt_card(info);
            return ret;
        }
    }
    rs485_transmit_disable(info);
    rs485_allow_interrupts();
    rs485_allow_interrupt_card(info);  
// REPLY:    
#ifdef __DEBUG_TIMING__
    rs485_gettime(&tm_start);
#endif
    rs485_gettime(&end_packet_out_tm);
    start_packet_inp_tm = end_packet_out_tm;
    tm = end_packet_out_tm;
    rs485_set_starttime();
    
    while( ( info->rs->status & ( SIGNAL_RECEIVED | DATA_READY ) ) == 0 ) {
        if( rs485_timeout( RBYTE_TIMEOUT ) ) {
            if( info->rs->status & ( SIGNAL_RECEIVED | DATA_READY ) ) break;	
            rs485_gettime(&tm);
            rs485_printtm_interval(&end_packet_out_tm, &tm);
            PDEBUG( "transmit: timeout! addr=%hx status=%hx info->rs->inbuffer.size=%zd info->rs->usCountByteInp=%zd\n",
                last_address, info->rs->status, info->rs->inbuffer_pas.size, info->rs->usCountByteInp );
            if( info->rs->status & RECEIVING ) ret = SignalEER;
            else ret = SignalDSC;
            return ret;
        }
        udelay( RBYTE_WAIT );
    }
//

#ifdef __DEBUG_TIMING__
    rs485_time_t tm_end;
    rs485_gettime(&tm_end);
    rs485_printtm_interval(&tm_start, &tm_end);
#endif
//
    if( info->rs->status & SIGNAL_RECEIVED ) {
        info->rs->status &= ~SIGNAL_RECEIVED;
        return signal_from_device;
    }
    if( info->rs->status & DATA_READY ) {
        info->rs->status &= ~DATA_READY;
        if( crc( info->rs->inbuffer_pas.cache, info->rs->inbuffer_pas.size+1 ) == 0 ) {
            ret = info->rs->inbuffer_pas.size;    // size of data without CRC
        } else {
            PDEBUG("addr=%hx Received packet with bad CRC!\n", last_address);
            ret = SignalEER;
        }
        return ret;
    }
    return -(ERSBASE+16);                 // algorithm. error
}                                    

static unsigned int counter = 0;

void receiv_data(RS485INFO * info) {
    unsigned short iir = 0;
    unsigned short csr = 0;
    counter = 0;
    while( ( ( csr = inb( info->base + UART_LSR ) ) & UART_LSR_DR ) != 0 && counter < 32 ) {
        counter++;

// debug
//        if( rs485_timeout(500) ) PDEBUG("counter=%d\n", counter);

        rs485_set_starttime();
        iir = inb( info->base );
        if( info->rs->status & SIGNAL_SENT ) { // Sent signal echo, check it !
            rs485_transmit_disable(info);
            info->rs->status |= DATA_READY;
            info->rs->status &= ~(PASSIVE_DATA|SIGNAL_SENT|RECEIVING);

            if(iir != SignalACK) PDEBUG("Bad echo! %hx != %hx\n", iir, SignalACK);
            if( (inb(info->base + UART_LSR) & UART_LSR_DR ) != 0 ) {
                PDEBUG("receive: addr=%hx echo received, fifo is not empty\n",
                          last_address);
            }
            return;
        }
        if( csr & UART_LSR_FE ) {
            PDEBUG( "receive: frame error, addr=%hx, status=%hx iir=%hx info->rs->usCountByteInp=%zd info->rs->inbuffer.size=%zd\n",
                last_address, info->rs->status, iir, info->rs->usCountByteInp, info->rs->inbuffer_pas.size );

            if( (inb(info->base + UART_LSR) & UART_LSR_DR ) != 0 ) {
                PDEBUG("addr=%hx fifo is not empty\n", last_address);
                cleanFIFO(info);
/*
                for(i=0; i<16; i++) {
                    if( (inb(info->base +UART_LSR) & UART_LSR_DR) != 0 ) {
                        PDEBUG("%hx\n",inb(info->base));
                    } else break;
                }
*/
            }
            return;
        }                            
        if( csr & UART_LSR_PE ) {
            info->rs->usCountByteInp = 0;
            info->rs->status &= ~RECEIVING;
        }
        if( (info->rs->status & RECEIVING) == 0  ) {  // First byte of the packet and PASSIVE mode !!!
            rs485_gettime(&start_packet_inp_tm);
            info->rs->status |= RECEIVING;
            info->rs->usCountByteInp = 0;
            if( iir >= (unsigned char)CACHE_MAXSIZE  ) { // Signal
                info->rs->status |= SIGNAL_RECEIVED;
                signal_from_device = iir;
                info->rs->status &= ~(PASSIVE_DATA|RECEIVING);

                if( (inb(info->base + UART_LSR) & UART_LSR_DR ) != 0 ) {
                    PDEBUG("receive: addr=%hx signal received, fifo is not empty\n",
                       last_address);
                    cleanFIFO(info);
                }
                return;
            }
            if( iir < PACKET_MAX_LENGTH ) {
                info->rs->inbuffer_pas.size = iir + 1;
                info->rs->inbuffer_pas.cache[0] = iir;
                info->rs->usCountByteInp = 1;
                continue;
            } else {
                PDEBUG( "receive: addr=%hx bad length, status=%hx\n", last_address, info->rs->status );
                continue;
            }
        }
        info->rs->inbuffer_pas.cache[info->rs->usCountByteInp] = iir;
        if( info->rs->usCountByteInp+1 > info->rs->inbuffer_pas.size ) {
// SEND ACK:
            outb_p( FIRST, info->base +UART_LCR );
            rs485_transmit_enable( info );
            info->rs->status |= SIGNAL_SENT;
            outb_p( SignalACK, info->base );

            if( (inb(info->base + UART_LSR) & UART_LSR_DR ) != 0 ) {
                PDEBUG("receive: addr=%hx send ACK, fifo is not empty\n", last_address);
                cleanFIFO(info);
            }
            return;
        }
        info->rs->usCountByteInp++;
    } // while-cycle
    return;
}

void rs485_blnumber_reset(RS485INFO *info) {
    int i;
    for(i=0; i<MAX_TASK_DEVICE; i++) {
        info->rs->device[i].number = 0;
        info->rs->device[i].round_count_prev = -1;
    }
    info->rs->inbuffer_pas.adr = 0;
    info->rs->inbuffer_act.adr = 0;
}

int rs485_open(struct inode *inode, struct file *pfile) {
    int err = 0;
    unsigned int minor = iminor(inode);
    if(minor >= COM_MAX) return -ENXIO;
    if((com_table[minor].flags & EXIST_FL) == 0) return -ENXIO;
    
    if(com_table[minor].use_mcua >= MAX_USE_STREAM ) return -ERSBUSY; // More than maximal shared users
    PDEBUG("%s: open device %d, %d\n", ADAPTER_ID, imajor(inode), iminor(inode));

    if(com_table[minor].use_mcua == 0) {    // First opening
        com_table[minor].rs = (struct RS *)kmalloc(sizeof(struct RS), GFP_KERNEL);
        if(!com_table[minor].rs) {
            return -ENOMEM;
        }
        rs485_blnumber_reset(&com_table[minor]);
        rs485_fifo_init(&com_table[minor]);
        
        rs485_get_reg(&com_table[minor]);
        err = rs485_reset(&com_table[minor]); 
        if(err < 0) {
            rs485_restore_reg(&com_table[minor]);
            return err;
        }
    }
    com_table[minor].use_mcua++;
    return 0;
}

int rs485_release(struct inode *inode, struct file *pfile) {
    RS485INFO *info = &(com_table[iminor(pfile->f_dentry->d_inode)]);

//    if( down_interruptible(&info->sem ) ) {
//        up(&info->sem);
//        return -EINTR;
//    }

    info->use_mcua--;

    if( info->use_mcua == 0 ) {
        rs485_stop_interrupt_card(info);        
        rs485_restore_reg(info);
        if(com_table[iminor(inode)].rs) {
            kfree(com_table[iminor(inode)].rs);
            com_table[iminor(inode)].rs = NULL;
        }

    }
//    udelay(SEM_WAIT);
//    up(&info->sem);
    return 0;
}

ssize_t rs485_write( struct file *fp, const char __user *buf, size_t size, loff_t *offset ) {
/*
 *  return:
 *  < 0 error;
 *  > 0x80 signal
 *  < 0x20 packet
 */
    RS485INFO    *info = &(com_table[iminor(fp->f_dentry->d_inode)]);
    unsigned long rest = 0;
    int i, ret = 0;

    if( down_interruptible( &info->sem ) ) {
        up(&info->sem);
        return -EINTR;
    }
    if( (size == 0) || (size > PACKET_MAX_LENGTH ) ) {
        PDEBUG( "rs485_write: size=%zd is invalid!\n", size);
        ret = -ERSFORMAT;
        goto OUT;
    }
    if( (ret=rs485_test_interface(info)) < 0 ) goto OUT;
    rest = copy_from_user( info->rs->outbuffer.cache, buf, size );
    if(rest != 0) {
        PDEBUG("rs485_write: copy_from_user rest=%ld\n", rest);
        ret = -EFAULT;
        goto OUT;
    }
    info->rs->outbuffer.size = size;
    crc( info->rs->outbuffer.cache, size );
    for(i=0; i<MAX_HOP; i++) {
        ret = start_transmitter( info );
        if( (ret != SignalEER) && (ret != SignalNAK) ) break;
        msleep(4);
    }
    if( ( ret > 0 ) && ( ret <= PACKET_MAX_LENGTH ) ) {
        rest = copy_to_user( (char*)buf, info->rs->inbuffer_pas.cache, ret);
        if( rest != 0 ) {
            PDEBUG( "rs485_write: copy_to_user rest=%ld\n", rest );
            ret = -EFAULT;
            goto OUT;
        }       
    }
OUT:
    udelay(SEM_WAIT);
    up(&info->sem );
    return ret;
}

ssize_t rs485_read( struct file *fp, char __user *buf,
            size_t count, loff_t *offset ) {
/* 
 * Read data fifo
 */    
    char count0;
    RS485INFO    *info = &(com_table[iminor(fp->f_dentry->d_inode)]);
    int ret = 0;

    if( down_interruptible(&info->sem)) {        // 2.4 ???
        up(&info->sem);
        return -EINTR;
    }

    ret = rs485_data_fifo_read(info, buf, 1);
    if( ret > 0 ) {
        get_user(count0, buf);    // buf in User space!!!
        if( count0 < (count-1) ) count = count0;
        else count--;
        ret = rs485_data_fifo_read( info, buf+1, count );
    }
//        udelay(SEM_WAIT);
    up(&info->sem);
    return ret;
}

#ifdef KERN_2_6_36
long rs485_ioctl(struct file *fp,
            unsigned int cmd, unsigned long arg) {
#else
int rs485_ioctl(struct inode *inode, struct file *fp,
            unsigned int cmd, unsigned long arg) {
#endif

    int err = SUCCESS;
    RS485INFO *info;
    unsigned long rest = 0;
    
    if(_IOC_TYPE(cmd) != RS_IOC_MAGIC) {        // 2.4 ???
        PDEBUG("ioctl_error!\n");
        return -ENOTTY;
    }
    if(_IOC_NR(cmd) > RS_IOC_MAXNR) {
        PDEBUG("ioctl_error!\n");
        return -ENOTTY;
    }
/*
    if(_IOC_DIR(cmd) & _IOC_READ) {
        acc = access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
    } else {
        acc = access_ok(VERIFY_READ, (void __user *)arg, _IOC_SIZE(cmd));
    }

    if(!acc) {
        PDEBUG("ioctl_error!\n");
        return -EFAULT;
    }
*/
    info = &(com_table[iminor(fp->f_dentry->d_inode)]);

    switch(cmd) {
        case RS_RESET:
            if( down_interruptible(&info->sem)) {
                up(&info->sem);
                return -EINTR;
            }
            err = rs485_reset(info);
            rs485_fifo_flush(info);
            udelay(SEM_WAIT);
            up(&info->sem);
            return err;

        case RS_FIFO_FLUSH:
            if( down_interruptible(&info->sem)) {
                up(&info->sem);
                return -EINTR;
            }
            rs485_fifo_flush(info);
            udelay(SEM_WAIT);
            up(&info->sem);
            return SUCCESS;

        case RS_SET_BAUDRATE:
            if( down_interruptible(&info->sem)) {
                up(&info->sem);
                return -EINTR;
            }
            err = rs485_test_device(info);
            if(err >= 0 ) {
                info->baudrate = (int)arg;
                err = rs485_set_baudrate(info);
            }
            udelay(SEM_WAIT);
            up(&info->sem);
            return err;

        case RS_GET_BAUDRATE:
            return info->baudrate;

        case RS_BLNUMBER_RESET:
            rs485_blnumber_reset(info);
            return SUCCESS;

        case RS_GET_VERS:
            rest = copy_to_user((char*)arg, RS485_UTS_RELEASE, sizeof(RS485_UTS_RELEASE));
            if(rest != 0) {
                PDEBUG("rs485_ioctl: get_vers: copy_to_user rest=%ld\n", rest);
                return -EFAULT;
            }
            return RS485_VERSION_CODE;

        case RS_GET_DATAFIFO_SIZE:
            return SIZE_STREAM_FIFO;

        default: break;
    }
    return -ERSNOIOCTL;
}

struct file_operations rs485_fops = {
    .owner        = THIS_MODULE,
    .open         = rs485_open,
    .release      = rs485_release,
    .read         = rs485_read,
    .write        = rs485_write,
#ifdef KERN_2_6_36
    .unlocked_ioctl = rs485_ioctl,
#else
    .ioctl        = rs485_ioctl,
#endif
};

MODULE_AUTHOR("Andre Zaitsev, Olga Voziakova (ovoz@sai.msu.ru)");
MODULE_DESCRIPTION("rs485 module");
#ifdef MODULE_LICENSE
    MODULE_LICENSE("GPL");
#endif

#ifdef KERN_2_6_18
static struct class *comdev_class;
#else
static struct class_simple *comdev_class;
#endif

static void rs485_release_resource(int port_number) {
    release_region(com_table[port_number].base, ADDRESS_REGION_SIZE);
    rs485_clean_irq(&com_table[port_number]);
    com_table[port_number].flags &= ~EXIST_FL;
    return;
}

void __init rs485_request_resource(int port_number) { // IO port & irq initialization
    if(!request_region(com_table[port_number].base, ADDRESS_REGION_SIZE, ADAPTER_ID)) {
        RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
                    ADAPTER_ID, (unsigned long) com_table[port_number].base);
        return;
    }
    if(rs485_request_irq(&com_table[port_number])) {
        release_region(com_table[port_number].base, ADDRESS_REGION_SIZE);
        return;
    }
    com_table[port_number].flags |= EXIST_FL;
}

__init int init_module(void) {
    int err = 0;
    int i;
    char version[]=RS485_UTS_RELEASE;
    for(i=0; i<COM_MAX; i++) {
        com_table[i].rs = NULL;
        com_table[i].use_mcua = 0;
        com_table[i].flags = 0;
        init_MUTEX(&com_table[i].sem);
        com_table[i].baudrate = baudrate[i];
        com_table[i].base = io[i];
        com_table[i].irq = irq[i];
        PDEBUG("init port %d: baudrate=%d base=%x irq=%d:\n",
            i+1, com_table[i].baudrate, com_table[i].base, com_table[i].irq);
    }

    if ( (major=register_chrdev (0, ADAPTER_ID, &rs485_fops)) <= 0) {
        printk ("rs485com: unable to get major\n");
        err = -EIO;
        goto out_chrdev;
    }
#ifdef KERN_2_6_18
    comdev_class = class_create(THIS_MODULE, ADAPTER_ID);
#else
    devfs_mk_dir("mcua");
    comdev_class = class_simple_create(THIS_MODULE, ADAPTER_ID);
#endif
    if(IS_ERR(comdev_class)) {
        err = PTR_ERR(comdev_class);
        goto out_class ;
    }
    RS_PRINT(KERN_INFO "%s module, version %s\n",
                 ADAPTER_ID, version);
    for(i=0; i<COM_MAX; i++) {
        rs485_request_resource(i);
        if((com_table[i].flags & EXIST_FL) == 0) continue;
#ifdef KERN_2_6_27
        device_create(comdev_class, NULL, MKDEV(major, i), NULL,
                    "rs485com%d", i);
#elif KERN_2_6_18
        class_device_create(comdev_class, NULL, MKDEV(major, i), NULL,
                    "rs485com%d", i);
#else
        class_simple_device_add(comdev_class, MKDEV(major, i), NULL,
                    "rs485com%d", i);
        devfs_mk_cdev( MKDEV(major, i), S_IFCHR | S_IRUGO | S_IWUGO,
                     "mcua%d", i);
#endif
        RS_PRINT(KERN_INFO "%s: install card on the base 0x%X and irq=%d\n",
            ADAPTER_ID, com_table[i].base, com_table[i].irq);
    }
    return 0;

    for(i=0; i<COM_MAX; i++) {
        if(!(com_table[i].flags & EXIST_FL)) continue;
#ifdef KERN_2_6_27
        device_destroy(comdev_class, MKDEV(major, i));
#elif KERN_2_6_18 
        class_device_destroy(comdev_class, MKDEV(major, i));
#else
        devfs_remove("mcua%d", i);
        class_simple_device_remove(MKDEV(major, i));
#endif
        rs485_release_resource(i);
    }


#ifdef KERN_2_6_18
    class_destroy(comdev_class);
out_class:
#else
    class_simple_destroy(comdev_class);
out_class:
    devfs_remove("mcua");
#endif
    unregister_chrdev(major, ADAPTER_ID);
out_chrdev:
    return err;
}

void cleanup_module(void){
    int i;
    for(i=0; i<COM_MAX; i++) {
        if(!(com_table[i].flags & EXIST_FL)) continue;
#ifdef KERN_2_6_27
	device_destroy(comdev_class, MKDEV(major, i));
#elif KERN_2_6_18
        class_device_destroy(comdev_class, MKDEV(major, i));
#else
        devfs_remove("mcua/%d",i);
        class_simple_device_remove(MKDEV(major, i));
#endif
        rs485_release_resource(i);
    }
#ifdef KERN_2_6_18
    class_destroy(comdev_class);
#else
    devfs_remove("mcua");
    class_simple_destroy(comdev_class);
#endif
    unregister_chrdev (major, ADAPTER_ID);
    return;
}
